import collections
from typing import Dict


class TokenUsageTrackerException(Exception):
    pass


class TokenUsageTracker:
    def __init__(self, budget=None):
        self.tracker = collections.Counter()
        self.budget = budget

    def clear(self):
        self.tracker = collections.Counter()
        return self

    def track_usage(self, usage: Dict):
        self.tracker = self.tracker + collections.Counter(usage)

        if self.budget is not None and self.tracker["total_tokens"] > self.budget:
            raise TokenUsageTrackerException(
                f"Token budget exceeded. Budget: {self.budget}, Usage: {dict(self.tracker)}"
            )
        return self.tracker



from __future__ import annotations

import json
import logging
import time
import uuid
from typing import List

from openai import OpenAI, Stream
from openai.types.chat.chat_completion_message import (
    ChatCompletionMessage,
    FunctionCall,
)

from actionweaver.actions.action import Action, ActionHandlers
from actionweaver.llms.openai.functions.functions import Functions
from actionweaver.utils import DEFAULT_ACTION_SCOPE
from actionweaver.utils.stream import get_first_element_and_iterator, merge_dicts
from actionweaver.utils.tokens import TokenUsageTracker


class OpenAIChatCompletionException(Exception):
    pass


class OpenAIChatCompletion:
    def __init__(self, model, token_usage_tracker=None, logger=None):
        self.model = model
        self.logger = logger or logging.getLogger(__name__)
        self.token_usage_tracker = token_usage_tracker or TokenUsageTracker()
        self.client = OpenAI()

        print(
            "\033[91mDeprecating soon. Please import the OpenAIChatCompletion class from actionweaver.llms.openai.tools.chat\033[0m"
        )

    def _invoke_function(
        self,
        call_id,
        messages,
        model,
        response_msg,
        function_call,
        functions,
        orch,
        action_handler,
    ):
        """Invoke the function, update the messages, returns functions argument for the next OpenAI API call or halt the function loop and return the response."""

        if isinstance(function_call, FunctionCall):
            function_call = function_call.model_dump()

        messages += [response_msg]

        name = function_call["name"]
        arguments = function_call["arguments"]

        if action_handler.contains(name):
            try:
                arguments = json.loads(function_call["arguments"])
            except json.decoder.JSONDecodeError as e:
                self.logger.error(
                    {
                        "message": "Parsing function call arguments from OpenAI response ",
                        "arguments": function_call["arguments"],
                        "timestamp": time.time(),
                        "model": model,
                        "call_id": call_id,
                    },
                    exc_info=True,
                )
                raise OpenAIChatCompletionException(e) from e

            # Invoke action
            function_response = action_handler[name](**arguments)
            stop = action_handler[name].stop
            messages += [
                {
                    "role": "function",
                    "name": name,
                    "content": str(function_response),
                }
            ]

            self.logger.debug(
                {
                    "message": "Action invoked and response received",
                    "action_name": name,
                    "action_arguments": arguments,
                    "action_response": function_response,
                    "timestamp": time.time(),
                    "call_id": call_id,
                    "stop": stop,
                }
            )

            expr = (
                orch[action_handler[name]]
                if orch[action_handler[name]] != DEFAULT_ACTION_SCOPE
                else orch[DEFAULT_ACTION_SCOPE]
            )

            return (
                Functions.from_expr(
                    expr,
                ),
                (stop, function_response),
            )
        else:
            unavailable_function_msg = f"{name} is not a valid function name, use one of the following: {', '.join([func['name'] for func in functions.functions])}"
            messages += [
                {
                    "role": "user",
                    "content": unavailable_function_msg,
                }
            ]
            self.logger.debug(
                {
                    "message": "Unavailable action",
                    "action_name": name,
                    "action_arguments": arguments,
                    "action_response": unavailable_function_msg,
                    "timestamp": time.time(),
                    "call_id": call_id,
                }
            )
            return functions, (False, None)

    def create(
        self,
        messages,
        *args,
        orch=None,
        stream=False,
        actions: List[Action] = [],
        **kwargs,
    ):
        """
        Invoke the OpenAI API with the provided messages and functions.

        Functions supplied to the API are determined by the specified scope and orchestration expression.

        Args:
            messages (list): List of message objects for the conversation.
            scope (str): Scope of the functions to be used.
            stream (bool): If True, returns a generator that yields the API responses.

        Returns:
            API response with generated output.
        """

        # Todo: pass call_id to the decorated method
        call_id = str(uuid.uuid4())

        model = kwargs.get("model", self.model)
        temperature = kwargs.get("temperature", 0.0)

        # Restart token usage tracker
        self.token_usage_tracker.clear()

        # TODO: add validation to orch,
        action_handler = ActionHandlers()

        if orch is None:
            orch = {}
        if DEFAULT_ACTION_SCOPE not in orch:
            orch[DEFAULT_ACTION_SCOPE] = actions

        buf = actions + list(orch.keys()) + list(orch.values())

        for element in buf:
            if isinstance(element, list):
                for e in element:
                    action_handler.name_to_action[e.name] = e
            elif isinstance(element, Action):
                action_handler.name_to_action[element.name] = element
        # default action scope if not following actions not specified
        for _, action in action_handler.name_to_action.items():
            if action not in orch:
                orch[action] = DEFAULT_ACTION_SCOPE

        self.logger.debug(
            {
                "message": "Creating new chat completion",
                "model": model,
                "input_messages": messages,
                "timestamp": time.time(),
                "call_id": call_id,
            }
        )

        response = None
        functions = Functions.from_expr(orch[DEFAULT_ACTION_SCOPE])

        while True:
            self.logger.debug(
                {
                    "message": "Calling OpenAI API",
                    "call_id": call_id,
                    "input_messages": messages,
                    "model": model,
                    "timestamp": time.time(),
                    **functions.to_arguments(),
                }
            )

            function_argument = functions.to_arguments()
            if function_argument["functions"]:
                api_response = self.client.chat.completions.create(
                    model=model,
                    temperature=temperature,
                    messages=messages,
                    stream=stream,
                    **function_argument,
                )
            else:
                api_response = self.client.chat.completions.create(
                    model=model,
                    temperature=temperature,
                    messages=messages,
                    stream=stream,
                )

            # logic to handle streaming API response
            if isinstance(api_response, Stream):
                first_element, iterator = get_first_element_and_iterator(api_response)

                if first_element.choices[0].delta.content is not None:
                    # if the first element is a message, return generator right away.
                    return iterator
                elif first_element.choices[0].delta.function_call:
                    # if the first element in generator is a function call, merge all the deltas.
                    l = list(iterator)

                    deltas = {}
                    for element in l:
                        delta = element.choices[0].delta.model_dump()
                        deltas = merge_dicts(deltas, delta)

                    first_element.choices[0].message = ChatCompletionMessage(**deltas)
                    api_response = first_element
                else:
                    raise OpenAIChatCompletionException(
                        f"Unsupported response from streaming API: {list(iterator)}"
                    )

            else:
                self.token_usage_tracker.track_usage(api_response.usage)

            self.logger.debug(
                {
                    "message": "Received response from OpenAI API",
                    "response": api_response,
                    "timestamp": time.time(),
                    "call_id": call_id,
                }
            )

            choice = api_response.choices[0]
            message = choice.message

            if message.function_call:
                functions, (stop, resp) = self._invoke_function(
                    call_id,
                    messages,
                    model,
                    message,
                    message.function_call,
                    functions,
                    orch,
                    action_handler,
                )
                if stop:
                    return resp
            elif message.content is not None:
                response = message.content

                # ignore last message in the function loop
                # messages += [{"role": "assistant", "content": message["content"]}]
                if choice.finish_reason == "stop":
                    """
                    Stop Reasons:

                    - Occurs when the API returns a message that is complete or is concluded by one of the stop sequences defined via the 'stop' parameter.

                    See https://platform.openai.com/docs/guides/gpt/chat-completions-api for details.
                    """
                    self.logger.debug(
                        {
                            "message": "Model decides to stop",
                            "model": model,
                            "timestamp": time.time(),
                            "call_id": call_id,
                        }
                    )

                    break
            else:
                raise OpenAIChatCompletionException(
                    f"Unsupported response from OpenAI api: {api_response}"
                )
        return response



import types
import uuid
from typing import Callable, List

from pydantic import BaseModel

from actionweaver.actions import Action
from actionweaver.actions.factories.function import action


def truncated_uuid4():
    return str(uuid.uuid4())[:8]


# def action_from_model_v0(
#     model: BaseModel,
#     stop=True,
#     name: str = None,
#     description: str = None,
#     decorators: List[Callable[..., None]] = [],
# ) -> Action:
#     def func(*args, **kwargs):
#         if args or len(kwargs) > 1:
#             raise ValueError(
#                 f"Invalid input: The method should not have positional arguments and should only accept one keyword argument: {model.__name__.lower()} of type {model.__name__}",
#                 f"args: {args}",
#                 f"kwargs: {kwargs}",
#             )

#         if kwargs:
#             (key,) = kwargs.keys()

#             # the only keyword argument should be the model name
#             if key != model.__name__.lower():
#                 raise ValueError(
#                     f"Invalid input: The method should accept a single keyword argument: {model.__name__.lower()}",
#                     f"instead: {key}",
#                 )

#             (value,) = kwargs.values()

#             try:
#                 ret = model(**value)
#                 return ret
#             except Exception as e:
#                 raise ValueError(
#                     f"Failed to parse with the {model.__name__.lower()} model",
#                     f"value: {value}",
#                 )

#     if name is None:
#         name = f"Create{model.__name__}"

#     if description is None:
#         description = f"Extract {model.__name__}"

#     func.__doc__ = description
#     func.__name__ = f"create_{model.__name__.lower()}_from_pydantic_model"

#     return action(
#         name=name,
#         pydantic_model=create_pydantic_model_from_function(
#             func, override_params={model.__name__.lower(): (model, ...)}
#         ),
#         stop=stop,
#         decorators=decorators,
#     )(func)


def action_from_model(
    model: BaseModel,
    stop=True,
    name: str = None,
    description: str = None,
    decorators: List[Callable[..., None]] = [],
) -> Action:
    def func(*args, **kwargs):
        if args:
            raise ValueError(
                f"Invalid input: The method should not have positional arguments and should only accept keyword arguments: {model.__name__.lower()} of type {model.__name__}",
                f"args: {args}",
                f"kwargs: {kwargs}",
            )

        return model.model_validate(kwargs)

    if name is None:
        name = f"Create{model.__name__}"

    if description is None:
        description = f"Extract {model.__name__}"

    func.__doc__ = description
    func.__name__ = f"create_{model.__name__.lower()}_from_pydantic_model"

    return action(
        name=name,
        pydantic_model=model,
        stop=stop,
        decorators=decorators,
    )(func)



from __future__ import annotations

import itertools
import json
import logging
import time
from collections import defaultdict
from typing import List, Optional, Union

from openai import AsyncOpenAI, OpenAI, Stream
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
)

import actionweaver.llms.loop_action as la
from actionweaver.actions.action import Action, ActionHandlers
from actionweaver.llms.exception_handler import ChatLoopInfo, ExceptionHandler
from actionweaver.llms.openai.tools.tools import Tools
from actionweaver.telemetry import traceable
from actionweaver.utils import DEFAULT_ACTION_SCOPE
from actionweaver.utils.stream import get_first_element_and_iterator, merge_dicts
from actionweaver.utils.tokens import TokenUsageTracker


class FunctionCallingLoopException(Exception):
    def __init__(self, message="", extra_info=None):
        super().__init__(message)
        self.extra_info = extra_info or {}

    def __str__(self):
        # Customize the string representation to include extra_info
        extra_info_str = ", ".join(
            f"{key}: {value}" for key, value in self.extra_info.items()
        )
        return f"{super().__str__()} | Additional Info: [{extra_info_str}]"


def invoke_tool(
    messages,
    model,
    response_msg,
    tool_calls,
    tools,
    orch,
    action_handler: ActionHandlers,
):
    messages += [response_msg]

    # if multiple type of functions are invoked, ignore orch and `stop` option
    called_tools = defaultdict(list)

    # TODO: right now invoke all tools iteratively, implement async tool invocation
    for tool_call in tool_calls:
        if isinstance(tool_call, ChatCompletionMessageToolCall):
            tool_call = tool_call.model_dump()

        name = tool_call["function"]["name"]
        arguments = tool_call["function"]["arguments"]

        if action_handler.contains(name):
            try:
                arguments = json.loads(tool_call["function"]["arguments"])
            except json.decoder.JSONDecodeError as e:
                raise FunctionCallingLoopException(
                    f"Failed to parse function call arguments from OpenAI response",
                    extra_info={
                        "arguments": tool_call["function"]["arguments"],
                        "timestamp": time.time(),
                        "model": model,
                    },
                ) from e

            # Invoke action
            tool_response = action_handler[name](**arguments)

            called_tools[name].append(tool_response)

            stop = action_handler[name].stop
            messages += [
                {
                    "tool_call_id": tool_call["id"],
                    "role": "tool",
                    "name": name,
                    "content": str(tool_response),
                },
            ]

        else:
            raise FunctionCallingLoopException(
                f"{name} is not a valid function name",
                extra_info={
                    "timestamp": time.time(),
                    "model": model,
                },
            )

    if len(called_tools) == 1:
        # Update new functions for next OpenAI api call
        name = list(called_tools.keys())[0]

        # use tools in orch[DEFAULT_ACTION_SCOPE] if expr is DEFAULT_ACTION_SCOPE
        expr = (
            orch[name]
            if orch[name] != DEFAULT_ACTION_SCOPE
            else orch[DEFAULT_ACTION_SCOPE]
        )
        return (
            Tools.from_expr(
                expr,
            ),
            (stop, called_tools[name]),
        )
    else:
        # if multiple type of functions are invoked, use the same set of tools next api call
        return (
            tools,
            (False, list(called_tools.values())),
        )


def handle_stream_response(api_response):
    first_element, iterator = get_first_element_and_iterator(api_response)

    if first_element.choices[0].delta.content is not None:
        # if the first element is a message, return generator right away.
        return iterator
    else:
        # if the first element is a tool call, merge all tool calls into first response and return it
        l = list(iterator)

        deltas = {}
        for element in l:
            delta = element.choices[0].delta.model_dump()
            deltas = merge_dicts(deltas, delta)

        chat_completion_message_tool_call = defaultdict(dict)
        for tool_delta in deltas["tool_calls"]:
            chat_completion_message_tool_call[tool_delta["index"]] = merge_dicts(
                chat_completion_message_tool_call[tool_delta["index"]],
                tool_delta,
            )
            tool_delta.pop("index")

        deltas["tool_calls"] = list(chat_completion_message_tool_call.values())

        # (HACK) Remove the 'function_call' field, otherwise calling the API will fail
        if "function_call" in deltas:
            del deltas["function_call"]

        first_element.choices[0].message = ChatCompletionMessage(**deltas)

        return first_element


def build_orch(actions: List[Action] = None, orch=None):
    action_handler = ActionHandlers()

    if orch is None:
        orch = {}
    if DEFAULT_ACTION_SCOPE not in orch:
        orch[DEFAULT_ACTION_SCOPE] = actions

    buf = actions + list(orch.values())
    for element in buf:
        if isinstance(element, list):
            for e in element:
                action_handler.name_to_action[e.name] = e
        elif isinstance(element, Action):
            action_handler.name_to_action[element.name] = element
    # default action scope if not following actions not specified
    for _, action in action_handler.name_to_action.items():
        if action.name not in orch:
            orch[action.name] = DEFAULT_ACTION_SCOPE

    return action_handler, orch


def validate_orch(orch):
    if orch is not None:
        for key in orch.keys():
            if not isinstance(key, str):
                raise FunctionCallingLoopException(
                    f"Orch keys must be action name (str), found {type(key)}"
                )


def create_chat_loop(original_create_method):
    def wrapper_for_logging(
        *args,
        logger: Optional[logging.Logger] = None,
        logging_name: Optional[str] = None,
        logging_metadata: Optional[dict] = None,
        logging_level=logging.INFO,
        exception_handler: ExceptionHandler = None,
        **kwargs,
    ):
        DEFAULT_LOGGING_NAME = "actionweaver_initial_chat_completion"

        def new_create(
            actions: List[Action] = [],
            orch=None,
            token_usage_tracker=None,
            *args,
            **kwargs,
        ):
            validate_orch(orch)

            chat_completion_create_method = original_create_method
            if logger:
                chat_completion_create_method = traceable(
                    name=(logging_name or DEFAULT_LOGGING_NAME)
                    + ".chat.completions.create",
                    logger=logger,
                    metadata=logging_metadata,
                    level=logging_level,
                )(original_create_method)

            if token_usage_tracker is None:
                token_usage_tracker = TokenUsageTracker()

            argument_check(*args, **kwargs)

            messages = kwargs.get("messages")
            model = kwargs.get("model")

            action_handler, orch = build_orch(actions, orch)

            tools = Tools.from_expr(orch[DEFAULT_ACTION_SCOPE])
            chat_loop_action = la.Unknown

            while True:

                try:
                    if bool(tools):
                        tools_argument = tools.to_arguments()
                        api_response = chat_completion_create_method(
                            *args,
                            **kwargs,
                            **tools_argument,
                        )
                    else:
                        api_response = chat_completion_create_method(
                            *args,
                            **kwargs,
                        )

                    chat_loop_action = handle_response(
                        api_response,
                        token_usage_tracker,
                        messages,
                        model,
                        tools,
                        orch,
                        action_handler,
                        logger,
                    )
                except Exception as e:
                    if exception_handler:
                        chat_loop_action = exception_handler.handle_exception(
                            e,
                            ChatLoopInfo(
                                context={
                                    "response": api_response,
                                    "tools": tools,
                                    "messages": messages,
                                    "model": model,
                                    "orch": orch,
                                }
                            ),
                        )

                    else:
                        raise e

                if isinstance(chat_loop_action, la.ReturnRightAway):
                    return chat_loop_action.content
                elif isinstance(chat_loop_action, la.Continue):
                    tools = chat_loop_action.functions
                else:
                    raise FunctionCallingLoopException(
                        f"Unsupported chat loop action: {chat_loop_action}"
                    )

        if logger:
            return traceable(
                name=logging_name or DEFAULT_LOGGING_NAME,
                logger=logger,
                metadata=logging_metadata,
                level=logging_level,
            )(new_create)(*args, **kwargs)
        else:
            return new_create(*args, **kwargs)

    return wrapper_for_logging


def argument_check(
    *args,
    **kwargs,
):
    if "messages" not in kwargs:
        raise FunctionCallingLoopException(
            "messages keyword argument is required for chat completion"
        )
    if "model" not in kwargs:
        raise FunctionCallingLoopException(
            "model keyword argument is required for chat completion"
        )

    if "tools" in kwargs:
        raise FunctionCallingLoopException(
            "tools keyword argument is not allowed for this method, use actions instead"
        )

    if "tool_choice" in kwargs:
        raise FunctionCallingLoopException(
            "tool_choice keyword argument is not allowed for this method, use actions instead"
        )


def handle_response(
    api_response,
    token_usage_tracker,
    messages,
    model,
    tools,
    orch,
    action_handler,
    logger=None,
) -> la.LoopAction:

    # logic to handle streaming API response
    if isinstance(api_response, Stream):
        api_response = handle_stream_response(api_response)

        if isinstance(api_response, itertools._tee):
            # if it's a tee object, return right away
            return la.ReturnRightAway(content=api_response)
    else:
        token_usage_tracker.track_usage(api_response.usage)

    choice = api_response.choices[0]
    message = choice.message

    if message.tool_calls:
        tools, (stop, resp) = invoke_tool(
            messages,
            model,
            message,
            message.tool_calls,
            tools,
            orch,
            action_handler,
        )
        if stop:
            return la.ReturnRightAway(content=resp)
        else:
            return la.Continue(functions=tools)
    elif message.content is not None:

        # ignore last message in the function loop
        # messages += [{"role": "assistant", "content": message["content"]}]
        if choice.finish_reason == "stop":
            """
            Stop Reasons:

            - Occurs when the API returns a message that is complete or is concluded by one of the stop sequences defined via the 'stop' parameter.

            See https://platform.openai.com/docs/guides/gpt/chat-completions-api for details.
            """

            return la.ReturnRightAway(content=api_response)
    else:
        raise FunctionCallingLoopException(
            f"Unsupported response from OpenAI api: {api_response}"
        )



from actionweaver import action


class LangChainTools:
    def verify_lib_installed(self):
        try:
            import langchain
        except ImportError:
            raise ImportError(
                "`langchain` package not found, please run `pip install langchain`"
            )

    @action(name="GoogleSearch")
    def search(self, query: str):
        """
        Perform a Google search and return query results with titles and links.

        :param query: The search query to be used for the Google search.
        """
        from langchain.utilities import GoogleSearchAPIWrapper

        search = GoogleSearchAPIWrapper()
        res = search.results(query, 10)
        formatted_data = ""

        # Iterate through the data and append each item to the formatted_data string
        for idx, item in enumerate(res):
            formatted_data += f"({idx}) {item['title']}: {item['snippet']}\n"
            formatted_data += f"[Source]: {item['link']}\n\n"

        return formatted_data



from abc import ABC, abstractmethod
from typing import Any, Dict

from pydantic import BaseModel

from actionweaver.llms.loop_action import Continue as _Continue
from actionweaver.llms.loop_action import LoopAction
from actionweaver.llms.loop_action import ReturnRightAway as _ReturnRightAway
from actionweaver.llms.loop_action import Unknown as _Unknown


class ChatLoopInfo(BaseModel):
    context: Dict[str, Any]


class ExceptionAction(LoopAction):
    pass


class Continue(ExceptionAction, _Continue):
    pass


class Return(ExceptionAction, _ReturnRightAway):
    pass


class Unknown(ExceptionAction, _Unknown):
    pass


class ExceptionHandler(ABC):
    """Base class for exception handlers.

    This class provides a framework for handling exceptions within the function calling loop.
    """

    @abstractmethod
    def handle_exception(self, e: Exception, info: ChatLoopInfo) -> ExceptionAction:
        pass



# TODO: assume all actions are functions for now
from actionweaver.actions import Action


class ToolException(Exception):
    pass


class Tools:
    def __init__(self, tool_choice=None, tools=None) -> None:
        self.tools = tools
        self.tool_choice = tool_choice

    @classmethod
    def from_expr(cls, expr):

        if expr is None:
            return cls()
        elif isinstance(expr, Action):
            return cls(
                tool_choice={
                    "type": "function",
                    "function": {"name": expr.name},
                },
                tools=[
                    {
                        "type": "function",
                        "function": {
                            "name": expr.name,
                            "description": expr.description,
                            "parameters": expr.json_schema(),
                        },
                    }
                ],
            )
        elif isinstance(expr, list):
            return cls(
                tools=[
                    {
                        "type": "function",
                        "function": {
                            "name": action.name,
                            "description": action.description,
                            "parameters": action.json_schema(),
                        },
                    }
                    for action in expr
                ],
                tool_choice="auto",
            )
        else:
            raise ToolException(f"Invalid orchestration expression: {expr}")

    @staticmethod
    def from_action_to_json(action: Action):
        return {"type": "function", "function": action.get_function_details()}

    def to_arguments(self):
        return {
            "tools": self.tools,
            "tool_choice": self.tool_choice,
        }

    def __bool__(self):
        return bool(self.tools)



import contextvars
import functools
import inspect
import logging
import time
import traceback
import uuid
from typing import Any, Callable, Dict, Optional

_PARENT_RUN_ID = contextvars.ContextVar("_PARENT_RUN_ID", default=None)


def get_parent_run_id():
    return _PARENT_RUN_ID.get()


def _get_inputs(
    signature: inspect.Signature, *args: Any, **kwargs: Any
) -> Dict[str, Any]:
    """Return a dictionary of inputs from the function signature."""
    bound = signature.bind_partial(*args, **kwargs)
    bound.apply_defaults()
    arguments = dict(bound.arguments)
    arguments.pop("self", None)
    arguments.pop("cls", None)
    for param_name, param in signature.parameters.items():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            # Update with the **kwargs, and remove the original entry
            # This is to help flatten out keyword arguments
            if param_name in arguments:
                arguments.update(arguments[param_name])
                arguments.pop(param_name)

    return arguments


# inspired by langsmith.run_helpers.traceable
def traceable(
    name,
    logger,
    metadata: Optional[Dict] = None,
    level=logging.INFO,
) -> Callable:
    original_metadata = metadata or {}

    def decorator(func: Callable):
        @functools.wraps(func)
        def wrapper(
            *args: Any,
            logging_extra: Optional[Dict] = None,
            **kwargs: Any,
        ) -> Any:
            parent_run_id = _PARENT_RUN_ID.get()

            run_id = uuid.uuid4()

            logging_extra = logging_extra or {}
            metadata = original_metadata.copy()
            metadata.update(logging_extra)

            signature = inspect.signature(func)
            inputs = _get_inputs(signature, *args, **kwargs)

            _PARENT_RUN_ID.set(run_id)
            try:
                function_result = func(*args, **kwargs)
                logger.log(
                    level,
                    {
                        "name": name,
                        "inputs": inputs,
                        "outputs": function_result,
                        "parent_run_id": parent_run_id,
                        "run_id": run_id,
                        "timestamp": time.time(),
                        **metadata,
                    },
                )
            except Exception as e:
                stacktrace = traceback.format_exc()
                logger.log(
                    level,
                    {
                        "name": name,
                        "inputs": inputs,
                        "error": stacktrace,
                        "parent_run_id": parent_run_id,
                        "run_id": run_id,
                        "timestamp": time.time(),
                        **metadata,
                    },
                )
                raise e
            finally:
                _PARENT_RUN_ID.set(parent_run_id)
            return function_result

        return wrapper

    return decorator



from __future__ import annotations

import itertools
import json
import logging
import time
from collections import defaultdict
from typing import List, Optional, Union

from openai import AsyncOpenAI, OpenAI, Stream
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
)

from actionweaver.actions.action import Action, ActionHandlers
from actionweaver.llms.openai.tools.tools import Tools
from actionweaver.telemetry import traceable
from actionweaver.utils import DEFAULT_ACTION_SCOPE
from actionweaver.utils.stream import get_first_element_and_iterator, merge_dicts
from actionweaver.utils.tokens import TokenUsageTracker


class OpenAIChatCompletionException(Exception):
    def __init__(self, message="", extra_info=None):
        super().__init__(message)
        self.extra_info = extra_info or {}

    def __str__(self):
        # Customize the string representation to include extra_info
        extra_info_str = ", ".join(
            f"{key}: {value}" for key, value in self.extra_info.items()
        )
        return f"{super().__str__()} | Additional Info: [{extra_info_str}]"


class OpenAIChatCompletion:
    def __init__(self, model, token_usage_tracker=None, logger=None):
        self.model = model
        self.logger = logger or logging.getLogger(__name__)
        self.token_usage_tracker = token_usage_tracker or TokenUsageTracker()
        self.client = OpenAI()

    @staticmethod
    def _invoke_tool(
        messages,
        model,
        response_msg,
        tool_calls,
        tools,
        orch,
        action_handler: ActionHandlers,
    ):
        messages += [response_msg]

        # if multiple type of functions are invoked, ignore orch and `stop` option
        called_tools = defaultdict(list)

        # TODO: right now invoke all tools iteratively, implement async tool invocation
        for tool_call in tool_calls:
            if isinstance(tool_call, ChatCompletionMessageToolCall):
                tool_call = tool_call.model_dump()

            name = tool_call["function"]["name"]
            arguments = tool_call["function"]["arguments"]

            if action_handler.contains(name):
                try:
                    arguments = json.loads(tool_call["function"]["arguments"])
                except json.decoder.JSONDecodeError as e:
                    raise OpenAIChatCompletionException(
                        e,
                        extra_info={
                            "message": "Parsing function call arguments from OpenAI response ",
                            "arguments": tool_call["function"]["arguments"],
                            "timestamp": time.time(),
                            "model": model,
                        },
                    ) from e

                # Invoke action
                tool_response = action_handler[name](**arguments)

                called_tools[name].append(tool_response)

                stop = action_handler[name].stop
                messages += [
                    {
                        "tool_call_id": tool_call["id"],
                        "role": "tool",
                        "name": name,
                        "content": str(tool_response),
                    },
                ]

            else:
                # TODO: allow user to add callback for unavailable tool
                unavailable_tool_msg = f"{name} is not a valid tool name, use one of the following: {', '.join([tool['function']['name'] for tool in tools.tools])}"

                raise OpenAIChatCompletionException(unavailable_tool_msg)

        if len(called_tools) == 1:
            # Update new functions for next OpenAI api call
            name = list(called_tools.keys())[0]

            # use tools in orch[DEFAULT_ACTION_SCOPE] if expr is DEFAULT_ACTION_SCOPE
            expr = (
                orch[name]
                if orch[name] != DEFAULT_ACTION_SCOPE
                else orch[DEFAULT_ACTION_SCOPE]
            )
            return (
                Tools.from_expr(
                    expr,
                ),
                (stop, called_tools[name]),
            )
        else:
            # if multiple type of functions are invoked, use the same set of tools next api call
            return (
                tools,
                (False, list(called_tools.values())),
            )

    @staticmethod
    def _handle_stream_response(api_response):
        first_element, iterator = get_first_element_and_iterator(api_response)

        if first_element.choices[0].delta.content is not None:
            # if the first element is a message, return generator right away.
            return iterator
        else:
            # if the first element is a tool call, merge all tool calls into first response and return it
            l = list(iterator)

            deltas = {}
            for element in l:
                delta = element.choices[0].delta.model_dump()
                deltas = merge_dicts(deltas, delta)

            chat_completion_message_tool_call = defaultdict(dict)
            for tool_delta in deltas["tool_calls"]:
                chat_completion_message_tool_call[tool_delta["index"]] = merge_dicts(
                    chat_completion_message_tool_call[tool_delta["index"]],
                    tool_delta,
                )
                tool_delta.pop("index")

            deltas["tool_calls"] = list(chat_completion_message_tool_call.values())

            # (HACK) Remove the 'function_call' field, otherwise calling the API will fail
            if "function_call" in deltas:
                del deltas["function_call"]

            first_element.choices[0].message = ChatCompletionMessage(**deltas)

            return first_element

    @staticmethod
    def build_orch(actions: List[Action] = None, orch=None):
        action_handler = ActionHandlers()

        if orch is None:
            orch = {}
        if DEFAULT_ACTION_SCOPE not in orch:
            orch[DEFAULT_ACTION_SCOPE] = actions

        buf = actions + list(orch.values())
        for element in buf:
            if isinstance(element, list):
                for e in element:
                    action_handler.name_to_action[e.name] = e
            elif isinstance(element, Action):
                action_handler.name_to_action[element.name] = element
        # default action scope if not following actions not specified
        for _, action in action_handler.name_to_action.items():
            if action.name not in orch:
                orch[action.name] = DEFAULT_ACTION_SCOPE

        return action_handler, orch

    def create(
        self,
        orch=None,
        actions: List[Action] = [],
        *args,
        **kwargs,
    ):
        if "model" not in kwargs:
            kwargs["model"] = self.model

        return OpenAIChatCompletion.wrap_chat_completion_create(
            self.client.chat.completions.create
        )(
            actions=actions,
            orch=orch,
            logger=self.logger,
            token_usage_tracker=self.token_usage_tracker,
            *args,
            **kwargs,
        )

    @staticmethod
    def validate_orch(orch):
        if orch is not None:
            for key in orch.keys():
                if not isinstance(key, str):
                    raise OpenAIChatCompletionException(
                        f"Orch keys must be action name (str), found {type(key)}"
                    )

    @staticmethod
    def wrap_chat_completion_create(original_create_method):
        def wrapper_for_logging(
            *args,
            logger: Optional[logging.Logger] = None,
            logging_name: Optional[str] = None,
            logging_metadata: Optional[dict] = None,
            logging_level=logging.INFO,
            **kwargs,
        ):
            DEFAULT_LOGGING_NAME = "actionweaver_initial_chat_completion"

            def new_create(
                actions: List[Action] = [],
                orch=None,
                token_usage_tracker=None,
                *args,
                **kwargs,
            ):
                OpenAIChatCompletion.validate_orch(orch)

                chat_completion_create_method = original_create_method
                if logger:
                    chat_completion_create_method = traceable(
                        name=(logging_name or DEFAULT_LOGGING_NAME)
                        + ".chat.completions.create",
                        logger=logger,
                        metadata=logging_metadata,
                        level=logging_level,
                    )(original_create_method)

                if token_usage_tracker is None:
                    token_usage_tracker = TokenUsageTracker()

                messages = kwargs.get("messages")
                if messages is None:
                    raise OpenAIChatCompletionException(
                        "messages keyword argument is required for chat completion"
                    )
                model = kwargs.get("model")
                if model is None:
                    raise OpenAIChatCompletionException(
                        "model keyword argument is required for chat completion"
                    )

                action_handler, orch = OpenAIChatCompletion.build_orch(actions, orch)
                response = None

                tools = Tools.from_expr(orch[DEFAULT_ACTION_SCOPE])

                while True:
                    tools_argument = tools.to_arguments()
                    if tools_argument["tools"]:
                        api_response = chat_completion_create_method(
                            *args,
                            **kwargs,
                            **tools_argument,
                        )
                    else:
                        api_response = chat_completion_create_method(
                            *args,
                            **kwargs,
                        )

                    # logic to handle streaming API response
                    if isinstance(api_response, Stream):
                        api_response = OpenAIChatCompletion._handle_stream_response(
                            api_response
                        )

                        if isinstance(api_response, itertools._tee):
                            # if it's a tee object, return right away
                            return api_response
                    else:
                        token_usage_tracker.track_usage(api_response.usage)

                    choice = api_response.choices[0]
                    message = choice.message

                    if message.tool_calls:
                        tools, (stop, resp) = OpenAIChatCompletion._invoke_tool(
                            messages,
                            model,
                            message,
                            message.tool_calls,
                            tools,
                            orch,
                            action_handler,
                        )
                        if stop:
                            return resp
                    elif message.content is not None:
                        response = api_response

                        # ignore last message in the function loop
                        # messages += [{"role": "assistant", "content": message["content"]}]
                        if choice.finish_reason == "stop":
                            """
                            Stop Reasons:

                            - Occurs when the API returns a message that is complete or is concluded by one of the stop sequences defined via the 'stop' parameter.

                            See https://platform.openai.com/docs/guides/gpt/chat-completions-api for details.
                            """

                            break
                    else:
                        raise OpenAIChatCompletionException(
                            f"Unsupported response from OpenAI api: {api_response}"
                        )
                return response

            if logger:
                return traceable(
                    name=logging_name or DEFAULT_LOGGING_NAME,
                    logger=logger,
                    metadata=logging_metadata,
                    level=logging_level,
                )(new_create)(*args, **kwargs)
            else:
                return new_create(*args, **kwargs)

        return wrapper_for_logging

    @staticmethod
    def patch(client: Union[OpenAI, AsyncOpenAI]):
        if isinstance(client, AsyncOpenAI):
            raise NotImplementedError(
                "AsyncOpenAI client is not supported for patching yet."
            )

        client.chat.completions.create = (
            OpenAIChatCompletion.wrap_chat_completion_create(
                client.chat.completions.create
            )
        )
        return client



from itertools import tee

from openai.types.chat.chat_completion_chunk import ChatCompletionChunk


def get_first_element_and_iterator(iterator):
    # Create two copies of the iterator
    iter1, iter2 = tee(iterator, 2)

    first_element = next(iter1)
    return first_element, iter2


def merge_dicts(dict1, dict2):
    merged_dict = dict1.copy()
    for key, value in dict2.items():
        if key in merged_dict:
            if value is None:
                continue

            if isinstance(value, ChatCompletionChunk):
                value = value.model_dump()

            if type(value) == dict:
                merged_dict[key] = merge_dicts(merged_dict[key], value)
            elif type(value) == str:
                merged_dict[key] += value
            elif type(value) == int:
                merged_dict[key] += value
            elif type(value) == list:
                if merged_dict[key] is None:
                    merged_dict[key] = []
                merged_dict[key] += value
            else:
                raise Exception(f"Unsupported type {type(value)}")

        else:
            merged_dict[key] = value
    return merged_dict



from typing import Callable, List

from actionweaver import Action
from actionweaver.utils.pydantic_utils import create_pydantic_model_from_func


def action_from_tool(
    tool,
    name=None,
    description=None,
    decorators: List[Callable[..., None]] = [],
    stop=False,
):
    name = name or tool.name
    description = description or tool.description

    act = Action(
        name=name,
        function=tool._run,
        pydantic_model=create_pydantic_model_from_func(
            tool.name.upper(), tool._run, ignored_params=["run_manager"]
        ),
        stop=stop,
        decorators=decorators,
        description=description,
    )

    return act



import collections
from typing import List

from actionweaver.actions import Action
from actionweaver.actions.factories.function import action
from actionweaver.utils.pydantic_utils import create_pydantic_model_from_func


def combine(
    acts: List[Action], name: str = None, description: str = None, reducer=None
) -> Action:
    # TODO: make action has a default of None, so the mode can choose to use it or not
    if reducer is None:
        reducer = lambda l: "\n".join([str(e) for e in l])

    def func(*args, **kwargs):
        value_list = list(kwargs.values())
        if args:
            raise ValueError(
                "Invalid input: The method should not have positional arguments",
                f"args: {args}",
            )

        ans = []
        for a, i in zip(acts, value_list):
            try:
                ans += [a(**i)]
            except Exception as e:
                raise ValueError(
                    f"Failed to invoke {a.name} with the value of {i}: {e}",
                )

        return reducer(ans)

    # Build the name string using all action names from acts
    action_names = "_".join(a.name.lower() for a in acts)
    if name is None:
        name = f"combine_{action_names}"

    if description is None:
        description = ""

    func.__doc__ = description
    func.__name__ = name

    params = collections.OrderedDict()
    for act in acts:
        params[act.name] = (act.pydantic_model, ...)

    return action(
        name=name.title(),
        pydantic_model=create_pydantic_model_from_func(
            func.__name__.title(), func, override_params=params
        ),
    )(func)



from typing import Any

from pydantic import BaseModel


class LoopAction(BaseModel):
    pass


class ReturnRightAway(LoopAction):
    content: Any


class Unknown(LoopAction):
    pass


class Continue(LoopAction):
    functions: Any



# TODO: assume all actions are functions for now
from actionweaver.actions import Action


class ToolException(Exception):
    pass


class Tools:
    def __init__(self, tools=None) -> None:
        self.tools = tools

    @classmethod
    def from_expr(cls, expr):
        if expr is None:
            return cls()
        elif isinstance(expr, list):
            return cls(
                tools=expr,
            )
        else:
            raise ToolException(f"Invalid orchestration expression: {expr}")

    def to_arguments(self):
        return "\n".join(
            [
                f"""{a.name}: \n description: {a.description} \n params: {a.json_schema()}"""
                for a in self.tools
            ]
        )
